/*
 * Copyright (C) 2019-2022 The Kraken authors. All rights reserved.
 * Copyright (C) 2022-present The WebF authors. All rights reserved.
 */
#include "event_listener_map.h"

namespace webf {

EventListenerMap::EventListenerMap() {}

static bool AddListenerToVector(EventListenerVector* vector,
                                const std::shared_ptr<EventListener>& listener,
                                const std::shared_ptr<AddEventListenerOptions>& options,
                                RegisteredEventListener* registered_event_listener,
                                uint32_t* listener_count) {
  *registered_event_listener = RegisteredEventListener(listener, options);

  if (std::find(vector->begin(), vector->end(), *registered_event_listener) != vector->end()) {
    return false;  // Duplicate listener.
  }

  vector->push_back(*registered_event_listener);
  *listener_count = vector->size();
  return true;
}

static bool RemoveListenerFromVector(EventListenerVector* listener_vector,
                                     const std::shared_ptr<EventListener>& listener,
                                     const std::shared_ptr<EventListenerOptions>& options,
                                     size_t* index_of_removed_listener,
                                     RegisteredEventListener* registered_event_listener,
                                     uint32_t* listener_count) {
  // Do a manual search for the matching listener. It is not
  // possible to create a listener on the stack because of the
  // const on |listener|.
  auto it = std::find_if(listener_vector->begin(), listener_vector->end(),
                         [listener, options](const RegisteredEventListener& event_listener) -> bool {
                           return event_listener.Matches(listener, options);
                         });

  if (it == listener_vector->end()) {
    *index_of_removed_listener = -1;
    return false;
  }

  *registered_event_listener = *it;
  *index_of_removed_listener = it - listener_vector->begin();
  listener_vector->erase(it);
  *listener_count = listener_vector->size();

  return true;
}

bool EventListenerMap::Contains(const AtomicString& event_type) const {
  for (const auto& entry : entries_) {
    if (entry.first == event_type)
      return true;
  }
  return false;
}

bool EventListenerMap::ContainsCapturing(const AtomicString& event_type) const {
  for (const auto& entry : entries_) {
    if (entry.first == event_type) {
      for (const auto& event_listener : *entry.second) {
        if (event_listener.Capture())
          return true;
      }
      return false;
    }
  }
  return false;
}

void EventListenerMap::Clear() {
  entries_.clear();
}

bool EventListenerMap::Add(const AtomicString& event_type,
                           const std::shared_ptr<EventListener>& listener,
                           const std::shared_ptr<AddEventListenerOptions>& options,
                           RegisteredEventListener* registered_event_listener,
                           uint32_t* listener_count) {
  for (const auto& entry : entries_) {
    if (entry.first == event_type)
      return AddListenerToVector(entry.second.get(), listener, options, registered_event_listener, listener_count);
  }

  entries_.emplace_back(event_type, std::make_unique<EventListenerVector>());
  return AddListenerToVector(entries_.back().second.get(), listener, options, registered_event_listener,
                             listener_count);
}

bool EventListenerMap::Remove(const AtomicString& event_type,
                              const std::shared_ptr<EventListener>& listener,
                              const std::shared_ptr<EventListenerOptions>& options,
                              size_t* index_of_removed_listener,
                              RegisteredEventListener* registered_event_listener,
                              uint32_t* listener_count) {
  for (unsigned i = 0; i < entries_.size(); ++i) {
    if (entries_[i].first == event_type) {
      bool was_removed = RemoveListenerFromVector(entries_[i].second.get(), listener, options,
                                                  index_of_removed_listener, registered_event_listener, listener_count);
      if (entries_[i].second->empty()) {
        entries_.erase(entries_.begin() + i);
      }
      return was_removed;
    }
  }

  return false;
}

EventListenerVector* EventListenerMap::Find(const AtomicString& event_type) const {
  for (const auto& entry : entries_) {
    if (entry.first == event_type)
      return entry.second.get();
  }

  return nullptr;
}

void EventListenerMap::Trace(GCVisitor* visitor) const {
  for (const auto& entry : entries_) {
    for (auto& listener : *entry.second) {
      listener.Trace(visitor);
    }
  }
}

}  // namespace webf
